/*{{{  file header */

/*

    Meta-Environment - An environment for language prototyping.
    Copyright (C) 2000  Stichting Mathematisch Centrum, Amsterdam,
                        The Netherlands.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

*/
/*
  $Id: editor.c,v 1.4 2000/05/15 15:42:23 markvdb Exp $  
 */

/*
 * The version of gse is written by Mark van den Brand.
 */

/*}}}  */
/*{{{  includes */

#include <stdlib.h>
#include <assert.h>

#include "se.h"
#include "editor.h"
#include "path.h"
#include "sort.h"
#include "area.h"
#include "focus.h"
#include "tree.h"
#include "posinfo.h"

/*}}}  */

/*{{{  global variables */

ATermTable editorInstances;

/*}}}  */

/*{{{  static ATerm newEditorInstanceGivenLength(ATerm atLength) */

static ATerm newEditorInstanceGivenLength(ATerm atLength)
{
  ATerm focusList;

  int length = getLength(atLength);
  ATerm textTree = ATparse("w(\"\")");

  ATerm path = newPath();
  ATerm sort = newSort("invalid");
  ATerm st = newStart(1);
  ATerm area = newArea(st, atLength);
  ATerm focus = newFocus(path, sort, area);

  textTree = putTreeLength(textTree, length);
  focus = setDirty(focus);
  focusList = newFocusList(focus);

  return ATmake("editor(<term>,<term>)", textTree, focusList); 
}

/*}}}  */
/*{{{  static ATerm newEditorInstanceGivenTree(ATerm tree, ATerm focusList) */

static ATerm newEditorInstanceGivenTree(ATerm tree, ATerm focusList)
{
  return ATmake("editor(<term>,<term>)", tree, focusList);
}

/*}}}  */
/*{{{  static ATermList currentFocus(ATerm editor, int location) */

static ATermList currentFocus(ATerm editor, int location)
{ 
  ATerm tmpEditor;
  ATermList focuses = (ATermList)getFocuses(editor), newFocuses;
  ATerm editorTree = getTreeInEditor(editor);

  if(ATisEmpty(focuses)) {
    ATerm newFocus = createFocusGivenLocation(editorTree, location);
    return ATmakeList1(newFocus);
  }
  else {
    ATerm firstFocus = ATgetFirst(focuses);
    if(inFocusArea(firstFocus, location)) {
      return focuses;
    }
    else {
      focuses = ATgetNext(focuses);
      tmpEditor = newEditorInstanceGivenTree(editorTree, (ATerm)focuses),
      newFocuses = currentFocus(tmpEditor, location);
      newFocuses = ATinsert(newFocuses, firstFocus);
      return removeChildren(firstFocus, newFocuses);
    }
  }
}

/*}}}  */
/*{{{  static ATerm updateEditorWithInsertions(tree, focuses, loc, inslen) */

static ATerm updateEditorWithInsertions(ATerm tree, ATermList focuses, 
                                        int location, int insertionLength)
{
  ATerm newFocus;
  ATermList newFocuses = currentFocus(
                           newEditorInstanceGivenTree(tree, (ATerm)focuses),
                           location);
  ATerm newPath = getFocusPath(ATgetFirst(newFocuses));
  ATerm newTree = updatePosInfo(tree, newPath, insertionLength);
  newFocuses = updateFocus(newFocuses, location, insertionLength);
  newFocus = ATgetFirst(newFocuses);
  newFocuses = ATgetNext(newFocuses);
  newFocus = setDirty(newFocus);
  newFocuses = ATinsert(newFocuses, newFocus);

  return newEditorInstanceGivenTree(newTree, (ATerm)newFocuses);
}

/*}}}  */
/*{{{  static ATerm updateEditorWithDeletions(tree, focuses, location, cnt) */

static ATerm updateEditorWithDeletions(ATerm tree, ATermList focuses,
                                       int location, int count)
{
  return updateEditorWithInsertions(tree, focuses, location, -count);
}  

/*}}}  */
/*{{{  static ATerm moveFocusToTop(ATerm editor) */

static ATerm moveFocusToTop(ATerm editor)
{
  ATerm editorTree = getTreeInEditor(editor);
  ATerm path = newPath();  
  ATerm newFocus = createFocusGivenPath(editorTree, path);
  ATermList focuses = ATmakeList1(newFocus);

  return newEditorInstanceGivenTree(editorTree, (ATerm)focuses);
}

/*}}}  */

/*{{{  ATerm newEditorInstanceGivenText(char *text) */

ATerm newEditorInstanceGivenText(char *text)
{
  ATerm length = newLength(strlen(text));
  return newEditorInstanceGivenLength(length);
}

/*}}}  */
/*{{{  ATerm getTreeInEditor(ATerm editor) */

ATerm getTreeInEditor(ATerm editor)
{
  ATerm tree;

  if (ATmatch(editor, "editor(<term>,<term>)", &tree, NULL)) {
    return tree;
  }
  else {
    return NULL;
  }
}

/*}}}  */
/*{{{  ATerm getFocuses(ATerm editor) */

ATerm getFocuses(ATerm editor)
{
  ATerm focusList;

  if (ATmatch(editor, "editor(<term>,<term>)", NULL, &focusList)) {
    return focusList;
  }
  else {
    return NULL;
  }
}

/*}}}  */
/*{{{  ATerm putFocuses(ATerm editor, ATerm newFocusList) */

ATerm putFocuses(ATerm editor, ATerm newFocusList)
{
  ATerm oldFocusList, tree;

  if (ATmatch(editor, "editor(<term>,<term>)", &tree, &oldFocusList)) {
    return ATmake("editor(<term>,<term>)", tree, newFocusList);
  }
  else {
    return editor;
  }
}

/*}}}  */
/*{{{  ATerm replaceFocusByTree(ATerm editor, ATerm tree) */

ATerm replaceFocusByTree(ATerm editor, ATerm tree)
{
  ATermList focuses = (ATermList)getFocuses(editor);
  
  if(ATisEmpty(focuses)) {
    return editor;
  }
  else {
    ATerm editorTree = getTreeInEditor(editor);
    ATerm focus = ATgetFirst(focuses), newFocus;
    ATerm path, newTree, sort;

    focuses = ATgetNext(focuses);

    path = getFocusPath(focus);
    newTree = updateTerm(editorTree, path, tree); 
    newTree = annotateTreeWithPosInfo(newTree);
    sort = getTreeSort(tree);
    newFocus = setFocusSort(focus, sort);
    newFocus = setDirty(newFocus);
    
    focuses = ATinsert(focuses, newFocus);

    return newEditorInstanceGivenTree(newTree, (ATerm)focuses);
  }
}

/*}}}  */
/*{{{  ATerm makeTreeInvalidInEditor(ATerm editor) */

ATerm makeTreeInvalidInEditor(ATerm editor)
{
  ATermList focuses;
  ATerm topFocus, area, length, newEditor;

  editor = moveFocusToTop(editor);
  focuses = (ATermList)getFocuses(editor);
  topFocus = ATgetFirst(focuses);
  area = getFocusArea(topFocus);
  length = getLengthOfArea(area);
  newEditor = newEditorInstanceGivenLength(length);

  return newEditor;
}

/*}}}  */
/*{{{  ATerm insertCharsAtLocation(ATerm editor, int location, char *text) */

ATerm insertCharsAtLocation(ATerm editor, int location, char *text)
{
  ATerm editorTree, newFocus;
  ATermList focuses;
  int textLength = strlen(text);

  editor = makeTreeInvalidInEditor(editor);
  focuses = (ATermList)getFocuses(editor);
  editorTree = getTreeInEditor(editor);
  if(ATisEmpty(focuses)) {
    newFocus = createFocusGivenLocation(editorTree, location);
    focuses = ATmakeList1(newFocus);
  }

  return updateEditorWithInsertions(editorTree, focuses, location, textLength);
}

/*}}}  */
/*{{{  ATerm deleteCharsAtLocation(ATerm editor, int location, int count) */

ATerm deleteCharsAtLocation(ATerm editor, int location, int count)
{
  ATerm editorTree, newFocus;
  ATermList focuses;

  assert(count == 1);
  editor = makeTreeInvalidInEditor(editor);
  focuses = (ATermList)getFocuses(editor);
  editorTree = getTreeInEditor(editor);
  if(ATisEmpty(focuses)) {
    newFocus = createFocusGivenLocation(editorTree, location);
    focuses = ATmakeList1(newFocus);
  }

  return updateEditorWithDeletions(editorTree, focuses, location, count);
}

/*}}}  */
/*{{{  ATerm moveFocusInEditor(ATerm editor, ATerm move) */

ATerm moveFocusInEditor(ATerm editor, ATerm move)
{
  ATermList focuses = (ATermList)getFocuses(editor);

  if(!ATisEmpty(focuses)) {
    ATerm newEditor = applyMoveToFocus(editor, move);
    ATermList newFocuses = (ATermList)getFocuses(newEditor);
    ATerm firstFocus = ATgetFirst(newFocuses);
    ATerm sort = getFocusSort(firstFocus);
    return newEditor;
  }
  else {
    return moveFocusToTop(editor);
  }
}

/*}}}  */
